/****************************************************************************
 *
 *           Program : d.c
 *
 *	     Author  : Kim, Dongmin
 *
 *
 *  This is a program that calculate the volume of a d-dimensional hypersphere
 *
 *  of radous 1 of arbitrary dimension d, where 1 <= d <= 10
 *
 *  by using Monte Carlo technique.
 *
 *  The equation of d-dimensional hypersphere of radius 1 is :
 *
 *         1 = sqrt ( (x1^2) + (x2^2) + ... + (x^d) ) )
 *
 *  In this program, each processor generates the volume independently and
 *
 *  0 processor collects them and calculates the derivation.
 *
 *  For derivation, I used the formula
 *
 *  derivation = sqrt [ { sigma ( x^2) } / N - average ^ 2 ]
 *
 *  To complie : gcc pi-mc.c -o pi -L/usr/local/lib/rs6000/ch_p4 -lmpi -lm
 *
 *  To run     : mpirun -mr_machinefile machinelist -mr_np 4 pi
 *
 ****************************************************************************/

#include	<stdio.h>
#include	<mpi.h>
#include        <math.h>

#define MAX     10000000
#define G_start 0.0  /* global start point */
#define G_end   1.0  /* global end point */

/****************************************************************************
 *        Global Value
 ****************************************************************************/
double x[10];

/****************************************************************************
 *
 *         Procedure : f()
 *         Input     : int d
 *         Output    : double 
 *         Function  : calculate x0**2 + . . . + xd**2 and return this result.
 *
 ****************************************************************************/
double  f( d )
     int    d;
{
  int    i;
  double sum;

  sum = 0.;
  for ( i = 0 ; i < d ; i++ ) {
    sum = sum + x[i] * x[i];
  }
  return sqrt(sum);
}

/****************************************************************************
 *
 *         Procedure : hypersphere()
 *         Input     : int d, int seed
 *         Output    : double summation
 *         Function  : find the summation between 0 and 1 using Monte Carlo
 *                     technique.
 *                     In this program, MAX numbers are generated in each
 *                     processor.
 *
 ****************************************************************************/
double hypersphere( d, seed )
     int    d;
     long   seed;
{
  int    i, j;
  double sum;
  double temp;

  sum = 0.;
  srand48(seed);

  for ( i = 0 ; i < MAX ; i++ ) {
    for ( j = 0 ; j < d ; j++ ) {
      x[j] = (double) drand48();
    }

    temp = f(d);
    if ( temp <= 1 ) {
      sum = sum + 1.;
    }
  }

  return sum;

}

/****************************************************************************
 *
 *         Procedure : main()
 *         Function  : main routine of this program.
 *
 ****************************************************************************/
void main(argc, argv)
  int argc;
  char **argv;
{

  int    i;
  int    d;           /* dimension of the hypersphere */
  int    nprocs;      /* Number of processors */
  int    pnum;        /* Processor number starting from 0 */
  long   seed;        /* seed for random number */
  double local_start; /* local start point */
  double local_end;   /* local end point */
  double term;        /* the length of term */
  double ave;         /* average between local_start and local_end */
  double other;       /* result from other processor */
  double error;       /* difference */
  int    from;
  int    dim;
  double sum_2, sum, var, dev;

  MPI_Status  recv_status;

  MPI_Init(&argc, &argv);

  /* Who am I ? */
  MPI_Comm_rank(MPI_COMM_WORLD, &pnum);

  /* How many processors participate ? */
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

  /* calculate start point and end points */
  term = ( G_end - G_start ) / nprocs;

  local_start = G_start + pnum * term;
  local_end   = local_start + term;

  /* setup the seed using processor number */
  seed = (long)pnum;

  dim = 2;

  sum_2 = 0.;
  sum = 0.;

  /* calculate the sum */
  for ( d = 1 ; d <= 10 ; d++ ) {

    ave = hypersphere( d, seed ) / (double)MAX;

    ave = ave * dim;

    dim = dim * 2;

    printf("Processor : %2d  Dimension : %2d    volume : %.18f\n", pnum, d, ave);

    if ( pnum == 0 ) { /* 0 processor collects the sums from other processors */

      sum_2 = ave * ave;

      for ( from = 1 ; from < nprocs ; from++ ) {

	/* 0 processor receives the sums from the other processors */
	MPI_Recv( &other, 1, MPI_DOUBLE, from, 99, MPI_COMM_WORLD, &recv_status);

	ave = ave + other;

	sum_2 = sum_2 + other * other;
      }

      ave = ave / (double) nprocs;

      var = ( sum_2 / (double) nprocs ) - ave * ave;

      dev = sqrt(var);  

      printf("\n\n\n");
      printf("***************************************************************\n");
      printf("  Dimension : %d\n", d);
      printf("\n");
      printf("   Number of Processor(s) : %d\n", nprocs);
      printf("   volume ( by Monte Carlo ) = %.18f\n", ave);
      printf("   Variance   : %.18f\n", var );
      printf("   Derivation : %.18f\n", dev );
      printf("***************************************************************\n");

    }
    else { /* Processors which is not 0 send the sum to the 0 processor */
      MPI_Send( &ave, 1, MPI_DOUBLE, 0, 99, MPI_COMM_WORLD);
    }

  }
  MPI_Finalize();

}

